@@ -98,7 +98,9 @@ struct fallback_dispatch {
 	struct device_coords rel;
 
 	struct {
-		struct device_coords delta;
+		struct device_coords lo_res;
+		struct device_coords hi_res;
+		bool emulate_hi_res_wheel;
 		bool is_inhibited;
 	} wheel;
 
