@@ -471,6 +471,38 @@ START_TEST(range_prop_parser)
 }
 END_TEST
 
+START_TEST(boolean_prop_parser)
+{
+	struct parser_test_range {
+		char *tag;
+		bool success;
+		bool b;
+	} tests[] = {
+		{ "0", true, false },
+		{ "1", true, true },
+		{ "-1", false, false },
+		{ "2", false, false },
+		{ "abcd", false, false },
+		{ NULL, false, false }
+	};
+	int i;
+	bool success, b;
+
+	for (i = 0; tests[i].tag != NULL; i++) {
+		b = false;
+		success = parse_boolean_property(tests[i].tag, &b);
+		ck_assert(success == tests[i].success);
+		if (success)
+			ck_assert_int_eq(b, tests[i].b);
+		else
+			ck_assert_int_eq(b, false);
+	}
+
+	success = parse_boolean_property(NULL, NULL);
+	ck_assert(success == false);
+}
+END_TEST
+
 START_TEST(evcode_prop_parser)
 {
 	struct parser_test_tuple {
@@ -1234,31 +1266,6 @@ START_TEST(strstartswith_test)
 }
 END_TEST
 
-START_TEST(strsanitize_test)
-{
-	struct strsanitize_test {
-		const char *string;
-		const char *expected;
-	} tests[] = {
-		{ "foobar", "foobar" },
-		{ "", "" },
-		{ "%", "%%" },
-		{ "%%%%", "%%%%%%%%" },
-		{ "x %s", "x %%s" },
-		{ "x %", "x %%" },
-		{ "%sx", "%%sx" },
-		{ "%s%s", "%%s%%s" },
-		{ NULL, NULL },
-	};
-
-	for (struct strsanitize_test *t = tests; t->string; t++) {
-		char *sanitized = str_sanitize(t->string);
-		ck_assert_str_eq(sanitized, t->expected);
-		free(sanitized);
-	}
-}
-END_TEST
-
 START_TEST(list_test_insert)
 {
 	struct list_test {
@@ -1320,6 +1327,39 @@ START_TEST(list_test_append)
 }
 END_TEST
 
+START_TEST(list_test_foreach)
+{
+	struct list_test {
+		int val;
+		struct list node;
+	} tests[] = {
+		{ .val  = 1 },
+		{ .val  = 2 },
+		{ .val  = 3 },
+		{ .val  = 4 },
+	};
+	struct list_test *t;
+	struct list head;
+
+	list_init(&head);
+
+	ARRAY_FOR_EACH(tests, t) {
+		list_append(&head, &t->node);
+	}
+
+	/* Make sure both loop macros are a single line statement */
+	if (false)
+		list_for_each(t, &head, node) {
+			ck_abort_msg("We should not get here");
+		}
+
+	if (false)
+		list_for_each_safe(t, &head, node) {
+			ck_abort_msg("We should not get here");
+		}
+}
+END_TEST
+
 START_TEST(strverscmp_test)
 {
 	ck_assert_int_eq(libinput_strverscmp("", ""), 0);
@@ -1430,6 +1470,7 @@ litest_utils_suite(void)
 	tcase_add_test(tc, reliability_prop_parser);
 	tcase_add_test(tc, calibration_prop_parser);
 	tcase_add_test(tc, range_prop_parser);
+	tcase_add_test(tc, boolean_prop_parser);
 	tcase_add_test(tc, evcode_prop_parser);
 	tcase_add_test(tc, input_prop_parser);
 	tcase_add_test(tc, evdev_abs_parser);
@@ -1447,12 +1488,12 @@ litest_utils_suite(void)
 	tcase_add_test(tc, strstrip_test);
 	tcase_add_test(tc, strendswith_test);
 	tcase_add_test(tc, strstartswith_test);
-	tcase_add_test(tc, strsanitize_test);
 	tcase_add_test(tc, time_conversion);
 	tcase_add_test(tc, human_time);
 
 	tcase_add_test(tc, list_test_insert);
 	tcase_add_test(tc, list_test_append);
+	tcase_add_test(tc, list_test_foreach);
 	tcase_add_test(tc, strverscmp_test);
 	tcase_add_test(tc, streq_test);
 	tcase_add_test(tc, strneq_test);
